package ensembles;

import java.util.Iterator;

/**
 * Interface dfinissant un ensemble comme une collection non trie d'lments
 * sans doublons. Le fait que les lments sont considrs comme non tris
 * impliquera que la comparaison de deux ensembles ne devra pas prendre en
 * compte l'ordre (apparent) des lments.
 *
 * @author davidroussel
 */
public interface Ensemble<E> extends Iterable<E>
{
	/**
	 * Ajout d'un lment à un ensemble ssi celui ci n'est pas null et qu'il
	 * n'est pas djà prsent
	 *
	 * @param element l'lment à ajouter à l'ensemble (on considèrera que l'on
	 *            ne peut pas ajouter d'lment null)
	 * @return true si l'lment a pu être ajout à l'ensemble, false sinon ou
	 *         si l'on a tent d'insrer un lment null (auquel cas il n'est
	 *         pas insr)
	 */
	public abstract boolean ajout(E element);

	/**
	 * Retrait d'un lment de l'ensemble en utilisant le remove de l'itrateur
	 * fournit par {@link #iterator()}
	 *
	 * @param element l'lment à supprimer de l'ensemble
	 * @return true si l'lment tait prsent dans l'ensemble (au sens de la
	 *         comparaison profonde) et qu'il a t retir, false sinon
	 */
	public default boolean retrait(E element)
	{
		boolean res = false;
		Iterator<E> it = iterator();
		if (it != null)
		{
			while (it.hasNext())
			{
				if(it.next()==element) {
					res = true;
					it.remove();
				}
			}
		}
		
		return res;
	}

	/**
	 * Teste si l'ensemble est vide en utilisant l'itrateur ou bien le
	 * {@link #cardinal}
	 *
	 * @return renvoie true si l'ensemble ne contient aucun lment, false sinon
	 * @see ensembles.Ensemble#estVide()
	 * @note Attention, si l'on utilise cardinal dans estVide, il ne faut pas
	 *       utiliser estVide dans cardinal et vice versa.
	 */
	public default boolean estVide()
	{
		boolean res = true;
		Iterator<E> it = iterator();
		if (it != null)
		{
			if (it.hasNext())
			{
				res = false;
			}
		}
		return res;
	}

	/**
	 * Test d'appartenance d'un lment à l'ensemble en utilisant l'itrateur
	 * pour parcourir les lments
	 *
	 * @param element l'lment dont on doit tester l'appartenance
	 * @return true si l'lment est prsent dans l'ensemble (au sens de la
	 *         comparaison profonde), false sinon
	 */
	public default boolean contient(E element)
	{
		boolean res = false;
		Iterator<E> it = iterator();
		if (it != null)
		{
			while (it.hasNext())
			{
				if(it.next()==element) {
					res = true;
				}
			}
		}
		
		return res;
	}

	/**
	 * Test si ensemble est un sous-ensemble de l'ensemble courant. C'est à dire
	 * si l'ensemble courant contient tous les lments de l'ensemble pass en
	 * argument
	 *
	 * @note Si l'ensemble pass en argument est null il ne sera pas considr
	 *       comme contenu.
	 * @param ensemble l'ensemble dont on veut tester s'il est un sous ensemble
	 *            de l'ensemble courant
	 * @return true si ensemble est un sous-ensemble de l'ensemble courant,
	 *         false sinon. false si ensemble est null.
	 */
	public default boolean contient(Ensemble<E> ensemble)
	{
		boolean res = true;
		if(ensemble == null) {
			res = false;
		}
		else {
			for(E elt : ensemble) {
				if(!this.contient(elt)) {
					res = false;
				}
			}
		}
		
		
		return res;

	}

	/**
	 * Efface tous les lements de l'ensemble en utilisant le remove de
	 * l'itrateur fournit par {@link #iterator()}
	 */
	public default void efface()
	{
		Iterator<E> it = iterator();
		if (it != null)
		{
			while (it.hasNext())
			{
				it.next();
				it.remove();
			}
		}
	}

	/**
	 * Taille de l'ensemble en utilisant l'itrateur
	 *
	 * @return le nombre d'lments dans l'ensemble
	 * @see ensembles.Ensemble#cardinal() Attention : si l'on utilise estVide
	 *      dans cardinal, il ne faut pas utiliser cardinal dans estVide
	 * @note Cette mthode aura intrêt à être rimplmente dans les classes
	 *       filles qui utilisent des conteneurs pouvant donner leur taille
	 *       directement
	 */
	public default int cardinal()
	{
		int count = 0;

		Iterator<E> it = iterator();
		if (it != null)
		{
			while (it.hasNext())
			{
				it.next();
				count++;
			}
		}
		

		return count;
	}

	/**
	 * Union avec un autre ensemble : (this union ensemble).
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer une union
	 * @return un nouvel ensemble contenant l'union de l'ensemble courant et de
	 *         l'ensemble pass en argument
	 */
	public abstract Ensemble<E> union(Ensemble<E> ensemble);

	/**
	 * Implmentation de classe de l'union de deux ensemble dans un autre
	 * ensemble
	 *
	 * @param ens1 le premier ensemble
	 * @param ens2 le second ensemble
	 * @param res l'ensemble contenant l'union de ens1 et ens2
	 */
	public static <E> void union(Ensemble<E> ens1, Ensemble<E> ens2, Ensemble<E> res)
	{
		for(E elt : ens1) {
			res.ajout(elt);
		}
		for(E elt : ens2) {
			res.ajout(elt);
		}
	}

	/**
	 * Intersection avec un autre ensemble : (this inter ensemble).
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer une
	 *            intersection
	 * @return un nouvel ensemble contenant l'intersection de l'ensemble courant
	 *         et de l'ensemble pass en argument
	 */
	public abstract Ensemble<E> intersection(Ensemble<E> ensemble);

	/**
	 * Implmentation de classe de l'intersection de deux ensemble dans un autre
	 * ensemble
	 *
	 * @param ens1 le premier ensemble
	 * @param ens2 le second ensemble
	 * @param res l'ensemble contenant l'intersection de ens1 et ens2
	 */
	public static <E> void intersection(Ensemble<E> ens1, Ensemble<E> ens2, Ensemble<E> res)
	{
		for(E elt : ens1) {
			if(ens2.contient(elt)) {
				res.ajout(elt);
			}
		}
	}

	/**
	 * Complment avec un autre ensemble : (this - ensemble).
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer le complment
	 * @return un nouvel ensemble contenant uniquement les lments prsents
	 *         dans l'ensemble courant mais PAS dans l'ensemble pass en
	 *         argument
	 */
	public abstract Ensemble<E> complement(Ensemble<E> ensemble);

	/**
	 * Implmentation de classe du complment de deux ensembles dans un autre
	 * ensemble.
	 *
	 * @param ens1 le premier ensemble
	 * @param ens2 le second ensemble
	 * @param res l'ensemble contenant le complment de ens1 - ens2
	 */
	public static <E> void complement(Ensemble<E> ens1, Ensemble<E> ens2, Ensemble<E> res)
	{
		for(E elt : ens2) {
			if(!ens1.contient(elt)) {
				res.ajout(elt);
			}
		}
	}

	/**
	 * Diffrence symmtrique avec un autre ensemble : (this delta ensemble).
	 * L'ensemble correspondant à la diffrence symtrique contient les lment
	 * qui sont soit dans l'ensemble courant, soit dans l'autre ensemble mais
	 * pas dans les deux ensembles = (this - ensemble) union (ensemble - this)
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer une diffrence
	 *            symtrique
	 * @return un nouvel ensemble contenant la diffrence symmtrique de
	 *         l'ensemble courant et de l'ensemble pass en argument
	 * @see ensembles.Ensemble#difference(ensembles.Ensemble)
	 */
	public default Ensemble<E> difference(Ensemble<E> ensemble)
	{
		/*
		 * TODO Remplacer par l'implmentation en utilisant
		 * 	- Soit (A - B) ∪ (B - A)
		 * 	- Soit (A ∪ B) - (B ∩ A)
		 */
		return this.complement(ensemble).union(ensemble.complement(this));
	}

	/**
	 * Type des lements de l'ensemble
	 *
	 * @return une instance de la classe Class reprsentant le type des lements
	 *         de l'ensemble si celui ci n'est pas vide, ou bien null si
	 *         l'ensemble est vide.
	 * @note cette mthode sera utile dans l'implmentation de la mthode
	 *       {@link #equals(Object)} pour dterminer si deux ensembles ont le
	 *       même type d'lments
	 * @see ensembles.Ensemble#typeElements()
	 */
	@SuppressWarnings("unchecked")
	public default Class<E> typeElements()
	{
		Iterator<E> it = iterator();
		if (it != null)
		{
			if (it.hasNext())
			{
				return (Class<E>) it.next().getClass();
			}
		}

		return null;
	}

	// ------------------------------------------------------------------------
	// Mthodes à implmenter dfinies dans la classe Object
	// ------------------------------------------------------------------------

	/**
	 * Test d'galit entre deux ensembles
	 *
	 * @param o l'objet à comparer
	 * @return true si l'objet à comparer est un ensemble et qu'il contient les
	 *         mêmes lments (pas forcment dans le même ordre). Si les deux
	 *         ensembles sont vides on considère qu'ils seront gaux quel que
	 *         soit leur type de contenu (dans la mesure où l'on ne peut pas le
	 *         dterminer avec {@link ensembles.Ensemble#typeElements()}
	 * @note une interface ne peut pas implmenter par dfaut des mthodes
	 *       surcharges de la classe object (celles ci dpendant de l'tat
	 *       interne des objets, ce qui n'est pas le cas d'une interface)
	 */
	@Override
	public abstract boolean equals(Object o);

	/**
	 * Hashcode d'un ensemble. Le HashCode d'un ensemble doit être calcul comme
	 * tant la somme des hascodes de ses lments afin de ne pas tenir compte
	 * de l'ordre des lments dans la collection sous-jacente.
	 *
	 * @return le hashage d'un ensemble
	 * @note une interface ne peut pas implmenter par dfaut des mthodes
	 *       surcharges de la classe object (celles ci dpendant de l'tat
	 *       interne des objets, ce qui n'est pas le cas d'une interface)
	 */
	@Override
	public abstract int hashCode();

	/**
	 * Affichage des lement de l'ensemble sous la forme : par exemple pour un
	 * ensemble de 3 elts : "[elt1, elt2, elt3]" où eltn reprsente le toString
	 * du nième elt.
	 *
	 * @return une chaine de caractère reprsentant les lments de l'ensemble
	 *         spar par des virgules et encadr par des crochets
	 * @note une interface ne peut pas implmenter par dfaut des mthodes
	 *       surcharges de la classe object (celles ci dpendant de l'tat
	 *       interne des objets, ce qui n'est pas le cas d'une interface)
	 */
	@Override
	public abstract String toString();

	// ------------------------------------------------------------------------
	// Mthodes à implmenter dfinies dans l'interface Iterable<E>
	// ------------------------------------------------------------------------
	/**
	 * Factory method fournissant un itrateur sur l'ensemble
	 *
	 * @return un nouvel itrateur sur cet ensemble
	 */
	@Override
	public abstract Iterator<E> iterator();
}
