package ensembles;

import java.util.Iterator;

import tableaux.Tableau;

/**
 * Ensemble à base de tableaux
 *
 * @author davidroussel
 */
public class EnsembleTableau<E> extends EnsembleGenerique<E>
{
	/**
	 * Conteneur sous-jacent : un Tableau<E>
	 */
	protected Tableau<E> tableau;

	/**
	 * Constructeur par dfaut d'un ensemble à base de {@link tableaux.Tableau}
	 */
	public EnsembleTableau()
	{
		this.tableau = new Tableau<E>();
	}

	/**
	 * Constructeur de copie à partir d'un {@link Iterable}
	 *
	 * @param elements l'itrable dont on doit copier les lments
	 */
	public EnsembleTableau(Iterable<E> elements)
	{	
		this.tableau = new Tableau<E>();
		for(E elt : elements) {
			if(!this.contient(elt)) {
				this.tableau.ajouter(elt);
			}
		}
	}

	/**
	 * Ajout d'un lment à un ensemble ssi celui ci n'est pas null et qu'il
	 * n'est pas djà prsent
	 * Ce qui revient dans le cas prsent à ajouter un lment au tableau si
	 * celui ci n'y est pas djà prsent
	 *
	 * @param element l'lment à ajouter à l'ensemble (on considèrera que l'on
	 *            ne peut pas ajouter d'lment null)
	 * @return true si l'lment a pu être ajout à l'ensemble, false sinon ou
	 *         si l'on a tent d'insrer un lment null (auquel cas il n'est
	 *         pas insr)
	 * @see ensembles.EnsembleGenerique#ajout(java.lang.Object)
	 */
	@Override
	public boolean ajout(E element)
	{
		if(element==null) {
			return false;
		}
		else {
			if(this.contient(element)) {
				return false;
			}
			else {
				this.tableau.ajouter(element);
				return true;
			}
		}
	}

	/**
	 * Taille de l'ensemble : rimplmentation en utilisant les proprits du
	 * tableau sous-jacent plutôt que l'itrateur (amlioration de performances)
	 *
	 * @return le nombre d'lments dans l'ensemble
	 * @see ensembles.EnsembleGenerique#cardinal()
	 */
	@Override
	public int cardinal()
	{
		return this.tableau.taille();
	}

	/**
	 * Union avec un autre ensemble en rutilisant la mthode de classe union
	 * crite dans l'ensemble Gnrique (
	 * {@link ensembles.EnsembleGenerique#union(ensembles.Ensemble, ensembles.Ensemble, ensembles.Ensemble)}
	 * ) et un nouvel {@link ensemble.EnsembleTableau} pour stocker le rsultat.
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer une union
	 * @return un nouvel ensemble contenant l'union de l'ensemble courant et de
	 *         l'ensemble pass en argument
	 * @see ensembles.EnsembleGenerique#union(ensembles.Ensemble,
	 *      ensembles.Ensemble, ensembles.Ensemble)
	 */
	@Override
	public Ensemble<E> union(Ensemble<E> ensemble)
	{
		Ensemble<E> res = new EnsembleTableau<E>();
		Ensemble.union(ensemble, this, res);
		return res;
	}

	/**
	 * Intersection avec un autre ensemble en rutilisant la mthode de classe
	 * intersection crite dans l'ensemble Gnrique (
	 * {@link ensembles.EnsembleGenerique#intersection(ensembles.Ensemble, ensembles.Ensemble, ensembles.Ensemble)}
	 * ) et un nouvel {@link ensemble.EnsembleTableau} pour stocker le rsultat.
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer une
	 *            intersection
	 * @return un nouvel ensemble contenant l'intersection de l'ensemble courant
	 *         et de l'ensemble pass en argument
	 * @see ensembles.EnsembleGenerique#intersection(ensembles.Ensemble,
	 *      ensembles.Ensemble, ensembles.Ensemble)
	 */
	@Override
	public Ensemble<E> intersection(Ensemble<E> ensemble)
	{
		Ensemble<E> res = new EnsembleTableau<E>();
		Ensemble.intersection(ensemble, this, res);
		return res;
	}

	/**
	 * Complment avec un autre ensemble en rutilisant la mthode de classe
	 * complement crite dans l'ensemble Gnrique (
	 * {@link ensembles.EnsembleGenerique#complement(ensembles.Ensemble, ensembles.Ensemble, ensembles.Ensemble)}
	 * ) et un nouvel {@link ensemble.EnsembleTableau} pour stocker le rsultat.
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer le complment
	 * @return un nouvel ensemble contenant uniquement les lments prsents
	 *         dans l'ensemble courant mais PAS dans l'ensemble pass en
	 *         argument
	 * @see ensembles.EnsembleGenerique#complement(ensembles.Ensemble,
	 *      ensembles.Ensemble, ensembles.Ensemble)
	 */
	@Override
	public Ensemble<E> complement(Ensemble<E> ensemble)
	{
		Ensemble<E> res = new EnsembleTableau<E>();
		Ensemble.complement(ensemble, this, res);
		return res;
	}

	/**
	 * Factory method fournissant un itrateur sur l'ensemble en rutilisant
	 * l'itrateur du tableau sous-jacent
	 *
	 * @return un nouvel itrateur sur cet ensemble
	 * @see ensembles.EnsembleGenerique#iterator()
	 */
	@Override
	public Iterator<E> iterator()
	{
		return this.tableau.iterator();
	}
}
