package ensembles;

import java.util.Collection;
import java.util.Iterator;

/**
 * Implmentation gnrique partielle d'un ensemble tri sous forme de
 * dcorateur d'un ensemble ordinaire.
 *
 * @author davidroussel
 */
public abstract class EnsembleTriGenerique<E extends Comparable<E>>
extends EnsembleGenerique<E> implements EnsembleTri<E>
{
	/**
	 * Ensemble de base sous-jacent dcor par les ensembles tris.
	 */
	protected Ensemble<E> ensemble;

	/**
	 * Ajout d'un nouvel lment de manire  maintenir l'ensemble tri en
	 * utilisant la mthode {@link #insererAuRang(E element, int rang)} ssi
	 * l'lment peut tre insr dans cet ensemble tri
	 *
	 * @param element l'lment  ajouter de manire trie (on considrera que
	 *            l'on ne peut pas ajouter d'lment null)
	 * @return true si l'lment n'tait pas dj prsent dans l'ensemble, false
	 *         sinon ou si l'on a tent d'insrer un lment null (auquel cas il
	 *         n'est pas insr).
	 * @see ensembles.EnsembleListe#ajout(java.lang.Object)
	 */
	@Override
	public boolean ajout(E element)
	{
		if(element==null) {
			return false;
		}
		else {
			if(this.contient(element)) {
				return false;
			}
			else {
				int rang = this.rang(element);
				this.insererAuRang(element, rang);
				return true;
			}
		}
	}

	/**
	 * Insertion d'un nouvel lment au rang choisi en utilisant
	 * {@link #rang(E element)} pour calculer le rang d'insertion de l'lment
	 *
	 * @param element l'lement  insrer
	 * @param rang le rang o insrer cet lment
	 * @return true si l'lment a t insr au rang choisi, false si l'lement
	 *         n'a pas pu tre insr  cause d'un rang invalide
	 * @note On remarquera que la mthode ne teste pas au pralable l'existence
	 *       de l'lment  insrer dans l'ensemble car c'est la mthode
	 *       {@link #ajout(E)} qui s'en chargera
	 */
	protected abstract boolean insererAuRang(E element, int rang);

	/**
	 * Test d'galit d'un ensemble tri. Il est ncessaire de rimplmenter la
	 * comparaison avec un autre ensemble car l'ordre des lment aura son
	 * importance dans la comparaison ce qui n'tait pas le cas avec les
	 * ensembles non tris.
	 *
	 * @return true si l'objet obj est aussi un ensemble (pas forcment tri) et
	 *         qu'il contient exactement les mmes lments dans le mme ordre.
	 * @see ensembles.EnsembleGenerique#equals(java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object obj)
	{
		if(obj==null) {
			return false;
		}
		else if(obj == this) {
			return true;
		}
		else if(obj instanceof Ensemble<?>) {
			if(((Ensemble<?>)obj).typeElements() == this.typeElements()) {
				if(this.typeElements()==null) {
					return true;
				}
				else {
					boolean res = true;
					int i=0;
					for(E elt : (Ensemble<E>)obj) {
						if(!this.contient(elt)) {
							res = false;
						}
						else {
							if(this.rang(elt) != i) {
								res = false;
							}
						}
						i++;
					}
					/*if(obj instanceof EnsembleTri<?>) {
						for(E elt : (EnsembleTri<E>)obj) {
							if(((EnsembleTri<E>) obj).rang(elt) != this.rang(elt)) {
								res = false;
							}
						}
					}
					else {
						res = false;
					}*/
					return res;
				}
			}
			else {
				return false;
			}
		}
		else {
			return false;
		}
	}

	/**
	 * Code de hachage d'un ensemble tri. Il est ncessaire de rimplmenter le
	 * code de hachage pour les ensembles tris car on considrera que deux
	 * ensembles contenant les mme lments mais dans des ordres diffrents
	 * seront eux-mme diffrents. Il faut donc que la mthode hashCode prenne
	 * en compte l'ordre des lments (Comme dans les autres {@link Collection}
	 * d'ailleurs).
	 *
	 * @return le code de hachage de cet ensemble tri.
	 * @see listes.Liste#hashCode() ou tableaux.Tableau#hashCode() pour un
	 *      exemple de hashage utilisant l'ordre des lments
	 * @see ensembles.EnsembleGenerique#hashCode()
	 */
	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		for (E elt : this)
		{
			result = (prime * result) + (elt == null ? 0 : elt.hashCode());
		}
		return result;
	}

	/**
	 * Union avec un autre ensemble : reste semblable  l'union avec avec un
	 * ensemble non tri mais s'applique sur l'ensemble dcor {@link #ensemble}
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer une union
	 * @return un nouvel ensemble contenant l'union de l'ensemble courant et de
	 *         l'ensemble pass en argument
	 * @see ensembles.EnsembleGenerique#union(ensembles.Ensemble)
	 */
	@Override
	public Ensemble<E> union(Ensemble<E> autre)
	{
		return this.union(autre);
	}

	/**
	 * Intersection avec un autre ensemble : reste semblable  l'intersection
	 * avec avec un ensemble non tri mais s'applique sur l'ensemble dcor
	 * {@link #ensemble}
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer une
	 *            intersection
	 * @return un nouvel ensemble contenant l'intersection de l'ensemble courant
	 *         et de l'ensemble pass en argument
	 * @see ensembles.EnsembleGenerique#intersection(ensembles.Ensemble)
	 */
	@Override
	public Ensemble<E> intersection(Ensemble<E> autre)
	{
		return this.intersection(autre);
	}

	/**
	 * Complement avec un autre ensemble : reste semblable au complement avec
	 * avec un ensemble non tri mais s'applique sur l'ensemble dcor
	 * {@link #ensemble}
	 *
	 * @param ensemble l'autre ensemble avec lequel on veut crer un complement
	 * @return un nouvel ensemble contenant le complement de l'ensemble courant
	 *         et de l'ensemble pass en argument
	 * @see ensembles.EnsembleGenerique#complement(ensembles.Ensemble)
	 */
	@Override
	public Ensemble<E> complement(Ensemble<E> autre)
	{
		return this.complement(autre);
	}

	/**
	 * Factory method fournissant un itrateur sur l'ensemble en rutilisant
	 * l'itrateur de l'ensemble ordinaire sous-jacent.
	 *
	 * @return un nouvel itrateur sur cet ensemble
	 * @see ensembles.EnsembleGenerique#iterator()
	 */
	@Override
	
	public Iterator<E> iterator()
	{
		return ensemble.iterator();
	}
}
